from functools import wraps


def admin_action(short_description=None):
    """Wrapper for admin action.

    ``short_description (str)``
    Description of the function.

    Usage::

        from jnt_admin_tools.decorators import admin_action
        from django.contrib import admin
        from test_app.models import Bar

        @admin.register(Bar)
        class BarAdmin(admin.ModelAdmin):
            fields = ("name",)
            actions = ("update_objs",)

            @admin_action("Update objs")
            def update_objs(self, request, queryset):
                pass

    The wrapped method receives the action method.
    """

    def wrap(func):  # noqa: WPS430
        @wraps(func)
        def action_func(self, request, queryset):  # noqa: WPS430
            return func(self, request, queryset)

        action_func.short_description = short_description
        return action_func

    return wrap



from contextlib import ExitStack

from django.conf import settings


class ProfilingMiddleware:
    """Middleware for providing profiling info."""

    def __init__(self, get_response):
        """Initializing."""
        self._get_response = get_response
        self._profilers = settings.REQUEST_PROFILERS or []

    def __call__(self, request):
        """Handle request."""
        with ExitStack() as stack:
            for profiler in self._profilers:
                profiler.before_request(request, stack)

            response = self._get_response(request)
            for profiler in self._profilers:
                profiler.after_request(request, response)

        return response



def ensure_fields_saved(save_kwargs, fields):
    """Ensuring fields saving (extending "update_fields")."""
    if "update_fields" not in save_kwargs:
        return

    save_kwargs["update_fields"] = tuple(
        set(list(save_kwargs["update_fields"]) + fields),
    )



from collections.abc import Iterable

from django.forms import IntegerField, ValidationError

from jnt_django_toolbox.admin.widgets import (
    BitFieldReadOnlyWidget,
    BitFieldWidget,
)
from jnt_django_toolbox.models.fields.bit.types import BitHandler


class BitFieldFormField(IntegerField):
    """Form field for bit field."""

    readonly_widget = BitFieldReadOnlyWidget

    def __init__(self, choices=(), widget=BitFieldWidget, *args, **kwargs):
        """Initializing."""
        if isinstance(kwargs["initial"], int):
            iv = kwargs["initial"]
            iv_list = []
            for i in range(0, min(len(choices), 63)):
                if (1 << i) & iv > 0:
                    iv_list += [choices[i][0]]
            kwargs["initial"] = iv_list
        self.widget = widget
        super().__init__(widget=widget, *args, **kwargs)
        self.choices = choices
        self.widget.choices = choices

    def clean(self, value):
        """Validate value."""
        if not value:
            return 0

        # Assume an iterable which contains an item per flag that's enabled
        result = BitHandler(0, [key for key, value in self.choices])
        for k in value:
            try:
                setattr(result, str(k), True)
            except AttributeError:
                raise ValidationError("Unknown choice: {0}".format(k))
        return int(result)

    def has_changed(self, initial, data):
        """Check changed field."""
        changed = super().has_changed(initial, data)
        if changed:
            changed = self._has_changed(initial, data)

        return changed

    def _has_changed(self, initial, data) -> bool:
        """Check is changed field."""
        initial_value = 0

        if isinstance(initial, BitHandler):
            initial_value = int(initial)
        elif isinstance(initial, Iterable):
            initial_value = self.clean(initial)

        return self.clean(data) != initial_value



SECONDS_PER_MINUTE = 60
SECONDS_PER_HOUR = 60 * SECONDS_PER_MINUTE
SECONDS_PER_DAY = 24 * SECONDS_PER_HOUR
SECONDS_PER_WEEK = 7 * SECONDS_PER_DAY
SECONDS_PER_MONTH = 30 * SECONDS_PER_DAY
SECONDS_PER_YEAR = 30 * SECONDS_PER_MONTH



from typing import Optional


class BitFieldReadOnlyWidget:
    """Readonly widget for bit field."""

    empty_value = "-"

    def __init__(self, formfield) -> None:
        """Initializing."""
        self._formfield = formfield

    def render(self, name, value) -> Optional[str]:
        """Render html-string."""
        checked_values = [
            choice for choice, checked in list(value.items()) if checked
        ]

        if not checked_values:
            return self.empty_value

        return ", ".join(
            (
                str(value._labels[value.keys().index(choice)])
                for choice in checked_values
            ),
        )
